"""
Py-ART: The Python ARM Radar Toolkit
=====================================

"""

# print information on citing Py-ART, this message can be suppressed by
# setting the PYART_QUIET environment variable
_citation_text = """
## You are using the Python ARM Radar Toolkit (Py-ART), an open source
## library for working with weather radar data. Py-ART is partly
## supported by the U.S. Department of Energy as part of the Atmospheric
## Radiation Measurement (ARM) Climate Research Facility, an Office of
## Science user facility.
##
## If you use this software to prepare a publication, please cite:
##
##     JJ Helmus and SM Collis, JORS 2016, doi: 10.5334/jors.119
"""
from os import environ as _environ
if 'PYART_QUIET' not in _environ:
    print(_citation_text)

# Make sure that deprecation warnings get printed by default
import warnings as _warnings
_warnings.simplefilter("always", DeprecationWarning)

# print out helpful message if build fails or importing from source tree
from . import __check_build

# import subpackages
from . import core
from . import io
from . import correct
from . import graph
from . import map
from . import filters
from . import util
from . import testing
from . import config
from . import aux_io
from . import retrieve
from . import bridge

# root level functions
from .config import load_config
from ._debug_info import _debug_info

import os.path as _osp
import functools as _functools
from pkg_resources import DistributionNotFound, get_distribution
import sys as _sys

# Get the version
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = '0.0.0'

try:
    if _sys.version_info[:2] >= (3, 4):
        import importlib as _importlib
        specs = _importlib.util.find_spec('pytest')
        specs.loader.load_module()
    else:
        import imp as _imp
        _imp.find_module('pytest')
except (AttributeError, ImportError) as error:
    def _test(verbose=False):
        """
        This would invoke the Py-ART test suite, but pytest couldn't
        be imported so the test suite can not run.
        """
        raise ImportError("Could not load pytest. Unit tests not available."
                            " To run unit tests, please install pytest.")
else:
    def _test(verbose=False):
        """
        Invoke the Py-ART test suite.
        """
        import pytest
        pkg_dir = _osp.abspath(_osp.dirname(__file__))
        args = [pkg_dir, '--pyargs', 'pyart']
        if verbose:
            args.extend(['-v', '-s'])
        pytest.main(args=args)

# Do not use `test` as function name as this leads to a recursion problem
# with the pytest test suite.
#test = _test
#test_verbose = _functools.partial(test, verbose=True)
#test_verbose.__doc__ = test.__doc_

#vdict = get_version()
#__version__ = vdict['version']_



"""
Cython routines for specific differential phase retrievals.

"""

# Necessary and/or potential future improvements to this module:
#
# * High order finite difference schemes still need to be added. Note that the
#   added value of using a higher order scheme has yet to be determined and may
#   be fruitless in the end.


cimport cython


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def lowpass_maesaka_term(
        double[:, ::1] k, dr, finite_order, double[:, ::1] d2kdr2):
    """
    Compute the filter term.

    Compute the low-pass filter term found in Maesaka et al. (2012). This term
    represents the second-order derivative of the control variable k with
    respect to range. This subroutine does not currently support radars with
    variable range resolution.

    Parameters
    ----------
    k : 2D array of float64
        Control variable k defined in Maesaka et al. (2012). This variable is
        proportional to the square root of specific differential phase.
    dr : float
        The range resolution in meters.
    finite_order : str, 'low' or 'high'
        The finite difference accuracy to use when computing the second-order
        range derivative of the control variable k.
    d2kdr2 : 2D array of float64
        Second-order derivative of k with respect to range. Updated in place.

    """

    # define local variables
    cdef int nr, ng, g, r
    cdef double dr2

    nr = k.shape[0]
    ng = k.shape[1]
    dr2 = dr**2.

    # Use a low order finite difference scheme to compute the second-order
    # range derivative
    if finite_order == 'low':
        for r in range(nr):
            for g in range(ng):
                # For interior range gates, i.e. g = [1, ng-1]
                # use a centered difference scheme where p = 2.

                # When at ray boundaires, i.e., g = 0 or ng-1 use either a
                # forward or backward difference scheme where p = 1 depending
                # on which boundary

                # Computing --> d2k/dr2
                if g > 0 and g < ng-1:
                    d2kdr2[r, g] = (k[r, g+1] - 2.*k[r, g] + k[r, g-1]) / dr2
                elif g == 0:
                    d2kdr2[r, g] = (k[r, g] - 2.*k[r, g+1] + k[r, g+2]) / dr2
                else:
                    d2kdr2[r, g] = (k[r, g] - 2.*k[r, g-1] + k[r, g-2]) / dr2
    else:
        raise ValueError("Invalid finite_order")
    return


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def lowpass_maesaka_jac(
        double[:, ::1] d2kdr2, dr, double Clpf, finite_order,
        double[:, ::1] dJlpfdk):
    """
    Compute the Jacobian of the filter cost functional.

    Compute the Jacobian of the low-pass filter cost functional similar to
    equation (18) in Maesaka et al. (2012). This function does not currently
    support radars with variable range resolution.

    Parameters
    ----------
    d2kdr2 : 2D array of float64
       Second-order derivative of the control variable k with respect to range.
       The control variable k is proportional to the square root of specific
       differential phase.
    dr : float
       The range resolution in meters.
    Clpf : float
       The low-pass filter (radial smoothness) constraint weight.
    finite_order :  str, 'low' or 'high'
       The finite difference accuracy used to compute the second-order range
       derivative of the control variable k.
    dJlpfdk : 2D array of float64
       The Jacobian of the low-pass filter cost functional with respect to the
       control variable k.  Updated in place.

    """

    # Define local variables
    cdef int nr, ng, g, r
    cdef double dr2

    nr = d2kdr2.shape[0]
    ng = d2kdr2.shape[1]

    # The low-pass filter cost is defined as,
    # Jlpf = 0.5 * Clpf * sum[ (d2k/dr2)**2 ] ,
    # where the sum is over all range gates for all rays.
    dr2 = dr**2

    # The Jacobian of Jlpf when a low finite order has been used to compute the
    # second-order range derivative of the control variable k
    if finite_order == 'low':
        for r in range(nr):
            for g in range(ng):
                if g > 2 and g < ng - 3:
                    dJlpfdk[r, g] = Clpf * (
                        d2kdr2[r, g-1] - 2.*d2kdr2[r, g] +
                        d2kdr2[r, g+1]) / dr2
                elif g == 2:
                    dJlpfdk[r, g] = Clpf * (
                        d2kdr2[r, g-2] + d2kdr2[r, g-1] - 2.*d2kdr2[r, g] +
                        d2kdr2[r, g+1]) / dr2
                elif g == 1:
                    dJlpfdk[r, g] = Clpf * (
                        d2kdr2[r, g+1] - 2.*d2kdr2[r, g] -
                        2.*d2kdr2[r, g-1]) / dr2
                elif g == 0:
                    dJlpfdk[r, g] = Clpf * (
                        d2kdr2[r, g] + d2kdr2[r, g+1]) / dr2
                elif g == ng - 3:
                    dJlpfdk[r, g] = Clpf * (
                        d2kdr2[r, g+2] + d2kdr2[r, g+1] - 2.*d2kdr2[r, g] +
                        d2kdr2[r, g-1]) / dr2
                elif g == ng - 2:
                    dJlpfdk[r, g] = Clpf * (
                        d2kdr2[r, g-1] - 2.*d2kdr2[r, g] -
                        2.*d2kdr2[r, g+1]) / dr2
                else:
                    dJlpfdk[r, g] = Clpf * (
                        d2kdr2[r, g] + d2kdr2[r, g-1]) / dr2
    else:
        raise ValueError("Invalid finite_order")
    return
